/*
 * Copyright (C) 2009 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
#define LOG_TAG "CubicJni"

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <jni.h>
#include <JNIHelp.h>

#include <android/bitmap.h>
#include <cutils/log.h>


static void throwException(JNIEnv* env, const char* className, const char* message) {
	jclass exceptionClass = env->FindClass(className);
	env->ThrowNew(exceptionClass, message);
}

static void throwIllegalStateException(JNIEnv* env, char *message) {
	const char* className = "java/lang/IllegalStateException";
	throwException(env, className, message);
}

static unsigned blend565tab[33][2][256];

static void init() {
	for(unsigned alpha=0; alpha<=32; ++alpha) {
		unsigned f = alpha;

		for(unsigned i=0; i<256; ++i) {
			blend565tab[alpha][1][i] = (((i & 0xf8)*f) << 19) + (((i & 0x07)*f) <<  3) + (0x04008010 >> 1);
			blend565tab[alpha][0][i] = (((i & 0xe0)*f) >>  5) + (((i & 0x1f)*f) << 11);
		}
	}
}

static inline unsigned blend565(unsigned dst, unsigned src, unsigned alpha) {
	unsigned ialpha = 32-alpha;
	unsigned sum = blend565tab[alpha][0][src & 0xff] + blend565tab[ialpha][0][dst & 0xff] +
			blend565tab[alpha][1][src >> 8] + blend565tab[ialpha][1][dst >> 8];

	sum &= 0xf81f07e0;

	return (sum & 0xffff) + (sum >> 16);
}

///////////////////////////////////////////////////////////////////////////

struct TriPt {
	float x, y;
	float u, v;
};

typedef unsigned short Pixel16;

// Triangle setup based on Chris Hecker's GDM article on texture mapping.
// We define pixel centers on the display to be at integers.

void RenderTriangle(Pixel16 *dst, long dstpitch, Pixel16 *tex, TriPt *pt1, TriPt *pt2, TriPt *pt3, uint32_t s) {
	TriPt *pt, *pl, *pr;

	// Find top point

	if (pt1->y < pt2->y)		// 1 < 2
		if (pt1->y < pt3->y) {	// 1 < 2,3
			pt = pt1;
			pr = pt2;
			pl = pt3;
		} else {				// 3 < 1 < 2
			pt = pt3;
			pr = pt1;
			pl = pt2;
		}
	else						// 2 < 1
		if (pt2->y < pt3->y) {	// 2 < 1,3
			pt = pt2;
			pr = pt3;
			pl = pt1;
		} else {				// 3 < 2 < 1
			pt = pt3;
			pr = pt1;
			pl = pt2;
		}

	if (pl->y == pt->y && pt->y == pr->y)
		return;

	// Compute gradients

	float A;
	float one_over_A;
	float dudx, dvdx;
	float dudy, dvdy;
	uint32_t dudxi, dvdxi;

	A = (pt->y - pl->y) * (pr->x - pl->x) - (pt->x - pl->x) * (pr->y - pl->y);
	one_over_A = 1.0f / A;
	dudx = ((pr->u - pl->u) * (pt->y - pl->y) - (pt->u - pl->u) * (pr->y - pl->y)) * one_over_A;
	dvdx = ((pr->v - pl->v) * (pt->y - pl->y) - (pt->v - pl->v) * (pr->y - pl->y)) * one_over_A;
	dudy = ((pt->u - pl->u) * (pr->x - pl->x) - (pr->u - pl->u) * (pt->x - pl->x)) * one_over_A;
	dvdy = ((pt->v - pl->v) * (pr->x - pl->x) - (pr->v - pl->v) * (pt->x - pl->x)) * one_over_A;

	dudxi = (int)(dudx * 16777216.0f) << 3;
	dvdxi = (int)(dvdx * 16777216.0f) << 3;

	// Compute edge walking parameters

	float dxl1=0, dxr1=0, dul1=0, dvl1=0;
	float dxl2=0, dxr2=0, dul2=0, dvl2=0;

	// Compute left-edge interpolation parameters for first half.

	if (pl->y != pt->y) {
		dxl1 = (pl->x - pt->x) / (pl->y - pt->y);

		dul1 = dudy + dxl1 * dudx;
		dvl1 = dvdy + dxl1 * dvdx;
	}

	// Compute right-edge interpolation parameters for first half.

	if (pr->y != pt->y) {
		dxr1 = (pr->x - pt->x) / (pr->y - pt->y);
	}

	// Reject backfaces.

	if (dxl1 >= dxr1)
		return;

	// Compute third-edge interpolation parameters.

	if (pr->y != pl->y) {
		dxl2 = (pr->x - pl->x) / (pr->y - pl->y);

		dul2 = dudy + dxl2 * dudx;
		dvl2 = dvdy + dxl2 * dvdx;

		dxr2 = dxl2;
	}

	// Initialize parameters for first half.
	//
	// We place pixel centers at (x+0.5, y+0.5).

	float xl, xr, ul, vl, yf;
	int y, y1, y2;

	// y_start < y+0.5 to include pixel y.

	y = (int)floor(pt->y + 0.5f);
	yf = (y+0.5f) - pt->y;

	xl = pt->x + dxl1 * yf;
	xr = pt->x + dxr1 * yf;
	ul = pt->u + dul1 * yf;
	vl = pt->v + dvl1 * yf;

	// Initialize parameters for second half.

	float xl2, xr2, ul2, vl2;

	if (pl->y > pr->y) {		// Left edge is long side
		dxl2 = dxl1;
		dul2 = dul1;
		dvl2 = dvl1;

		y1 = (int)floor(pr->y + 0.5f);
		y2 = (int)floor(pl->y + 0.5f);

		yf = (y1+0.5f) - pr->y;

		// Step left edge.

		xl2 = xl + dxl1 * (y1 - y);
		ul2 = ul + dul1 * (y1 - y);
		vl2 = vl + dvl1 * (y1 - y);

		// Prestep right edge.

		xr2 = pr->x + dxr2 * yf;
	} else {					// Right edge is long side
		dxr2 = dxr1;

		y1 = (int)floor(pl->y + 0.5f);
		y2 = (int)floor(pr->y + 0.5f);

		yf = (y1+0.5f) - pl->y;

		// Prestep left edge.

		xl2 = pl->x + dxl2 * yf;
		ul2 = pl->u + dul2 * yf;
		vl2 = pl->v + dvl2 * yf;

		// Step right edge.

		xr2 = xr + dxr1 * (y1 - y);
	}

	// Rasterize!

	int u_correct=0, v_correct=0;

	if (dudx < 0)			u_correct = -1;
	else if (dudx > 0)		u_correct = 0;
	else if (dul1 < 0)		u_correct = -1;

	if (dvdx < 0)			v_correct = -1;
	else if (dvdx > 0)		v_correct = 0;
	else if (dvl1 < 0)		v_correct = -1;

	s = (s>>3) + (s>>7);

/*	if (y < 0)
		y = 0;

	if (y2 > 160)
		y2 = 160;*/

	dst += dstpitch * y;

	while(y < y2) {
		if (y == y1) {
			xl = xl2;
			xr = xr2;
			ul = ul2;
			vl = vl2;
			dxl1 = dxl2;
			dxr1 = dxr2;
			dul1 = dul2;
			dvl1 = dvl2;
		}

		int x1, x2;
		float xf;
		unsigned u, v;

		// x_left must be less than (x+0.5) to include pixel x.

		x1		= (int)floor(xl + 0.5f);
		x2		= (int)floor(xr + 0.5f);
		xf		= (x1+0.5f) - xl;

		u		= ((int)((ul + xf * dudx)*16777216.0f) << 3) + u_correct;
		v		= ((int)((vl + xf * dvdx)*16777216.0f) << 3) + v_correct;

		while(x1 < x2) {
			uint32_t A = tex[(u>>27) + (v>>27)*32];
			uint32_t B = dst[x1];

#ifdef RGB_1555
			//           0rrr rrgg gggb bbbb
			//           0111 1011 1101 1110 = 0x7bde
			//           0111 1100 0001 1111 = 0x7c1f : r b
			//           0100 0000 0001 0000 = 0x4010
			// 0000 1111 1000 0011 1110 0000 = 0x0f83e0
			//           0000 0011 1110 0000 = 0x03e0 : g
			//           0000 0010 0000 0000 = 0x0200
			//           0111 1100 0000 0000 = 0x7c00
			uint32_t p0 = (((A^B ) & 0x7bde)>>1) + (A&B);
			uint32_t p  = (((A^p0) & 0x7bde)>>1) + (A&B);
			dst[x1] = (uint16_t)(((((p&0x007c1f)*s + 0x004010) & 0x0f83e0) + (((p&0x0003e0)*s + 0x000200) & 0x007c00)) >> 5);
#else // RGB_565
#if 1
			dst[x1] = blend565(B, A, s);
#else
			//           rrrr rggg gggb bbbb
			//           1111 0111 1101 1110 = 0xf7de
			//           1111 1000 0001 1111 = 0xf81f : r b
			//           1000 0000 0001 0000 = 0x8010
			// 0001 1111 0000 0011 1110 0000 = 0x1f03e0
			//           0000 0111 1110 0000 = 0x07e0 : g
			//           0000 0100 0000 0000 = 0x0400
			//      0001 1111 1000 0000 0000 = 0x1f800
			uint32_t p0 = (((A^B ) & 0xf7de)>>1) + (A&B);
			uint32_t p  = (((A^p0) & 0xf7de)>>1) + (A&B);
			dst[x1] = (uint16_t)(((((p&0x00f81f)*s + 0x008010) & 0x1f03e0) >> 5) + ((((p&0x0007e0)*s + 0x000400) & 0x01f800) >> 6));
#endif
#endif
			++x1;

			u += dudxi;
			v += dvdxi;
		}

		dst += dstpitch;
		xl += dxl1;
		xr += dxr1;
		ul += dul1;
		vl += dvl1;

		++y;
	}
}

///////////////////////////////////////////////////////////////////////////

struct RECT {
	int left;
	int top;
	int right;
	int bottom;

	RECT(): left(0), top(0), right(0), bottom(0) {}
	RECT(int l, int t, int r, int b): left(l), top(t), right(r), bottom(b) {}

	void set(int l, int t, int r, int b) {
		left = l; top = t;
		right = r; bottom = b;
	}
};

static Pixel16 *tex;
static RECT rBounce;
static RECT rDirtyLast;
static int xpos, ypos, xvel, yvel;
static float cubeside;
static float vx[8][3];
static const int faces[6][4] = {
	{ 0, 1, 4, 5 },
	{ 2, 6, 3, 7 },
	{ 0, 4, 2, 6 },
	{ 1, 3, 5, 7 },
	{ 0, 2, 1, 3 },
	{ 4, 5, 6, 7 },
};

static void OffsetRect(RECT *rect, int xpos, int ypos)
{
	rect->left += xpos;
	rect->right += xpos;

	rect->top += ypos;
	rect->bottom += ypos;
}

static void UnionRect(RECT *dst, const RECT *src1, const RECT *src2)
{
	dst->set(src1->left, src1->top, src1->right, src1->bottom);

	if (src2->left < src2->right && src2->top < src2->bottom) {
		if (dst->left < dst->right && dst->top < dst->bottom) {
			if (dst->left > src2->left) dst->left = src2->left;
			if (dst->top > src2->top) dst->top = src2->top;
			if (dst->right < src2->right) dst->right = src2->right;
			if (dst->bottom < src2->bottom) dst->bottom = src2->bottom;
		} else {
			dst->left = src2->left;
			dst->top = src2->top;
			dst->right = src2->right;
			dst->bottom = src2->bottom;
		}
	}
}

static int VDRoundToInt(double x)
{
	return (int)floor(x + 0.5);
}

static void updateBitmap(const AndroidBitmapInfo *info, void *pixels)
{
	static float theta = 0.0;
	float xt, yt, zt;

	xt = sinf(theta) / 80.0f;
	yt = sinf(theta + 3.1415926535f * 2.0f / 3.0f) / 80.0f;
	zt = sinf(theta + 3.1415926535f * 4.0f / 3.0f) / 80.0f;
	theta = theta + 0.005f;

	xpos += xvel;
	ypos += yvel;

	if (xpos<rBounce.left) { xpos = rBounce.left; xvel = +1; }
	if (xpos>rBounce.right) { xpos = rBounce.right; xvel = -1; }
	if (ypos<rBounce.top) { ypos = rBounce.top; yvel = +1; }
	if (ypos>rBounce.bottom) { ypos = rBounce.bottom; yvel = -1; }

	RECT rDirty = RECT(0x7fffffff, 0x7fffffff, -1, -1);
	RECT rDirty2;

	for(int i=0; i<8; i++) {
		float x0 = vx[i][0];
		float y0 = vx[i][1];
		float z0 = vx[i][2];

		float x1 = x0 * cosf(zt) - y0 * sinf(zt);
		float y1 = x0 * sinf(zt) + y0 * cosf(zt);
		float z1 = z0;

		float x2 = x1 * cosf(yt) - z1 * sinf(yt);
		float y2 = y1;
		float z2 = x1 * sinf(yt) + z1 * cosf(yt);

		float x3 = x2;
		float y3 = y2 * cosf(xt) - z2 * sinf(xt);
		float z3 = y2 * sinf(xt) + z2 * cosf(xt);

		vx[i][0] = x3;
		vx[i][1] = y3;
		vx[i][2] = z3;

		int ix1 = (int)floor(x3);
		int iy1 = (int)floor(y3);

		if (rDirty.left   > ix1) rDirty.left   = ix1;
		if (rDirty.right  < ix1) rDirty.right  = ix1;
		if (rDirty.top    > iy1) rDirty.top    = iy1;
		if (rDirty.bottom < iy1) rDirty.bottom = iy1;
	}

	OffsetRect(&rDirty, xpos, ypos);
	UnionRect(&rDirty2, &rDirty, &rDirtyLast);
	rDirtyLast = rDirty;

	++rDirty2.right;
	++rDirty2.bottom;

//	GdiFlush();

//	g_vbAboutDst.BitBlt(rDirty2.left, rDirty2.top, &g_vbAboutSrc, rDirty2.left, rDirty2.top,
//		rDirty2.right+1-rDirty2.left, rDirty2.bottom+1-rDirty2.top);

	TriPt v[4];

	v[0].u =  0;	v[0].v = 0;
	v[1].u =  0;	v[1].v = 32;
	v[2].u = 32;	v[2].v = 0;
	v[3].u = 32;	v[3].v = 32;

	const float lightdir[3]={
		0.57735026918962576450914878050196f / (cubeside*cubeside),
		0.57735026918962576450914878050196f / (cubeside*cubeside),
		0.57735026918962576450914878050196f / (cubeside*cubeside),
	};

	for(int f=0; f<6; f++) {
		const int f0 = faces[f][0];
		const int f1 = faces[f][1];
		const int f2 = faces[f][2];
		const int f3 = faces[f][3];

		v[0].x = vx[f0][0] + xpos;
		v[0].y = vx[f0][1] + ypos;
		v[1].x = vx[f1][0] + xpos;
		v[1].y = vx[f1][1] + ypos;
		v[2].x = vx[f2][0] + xpos;
		v[2].y = vx[f2][1] + ypos;
		v[3].x = vx[f3][0] + xpos;
		v[3].y = vx[f3][1] + ypos;

		const float side0[3]={
			vx[f1][0] - vx[f0][0],
			vx[f1][1] - vx[f0][1],
			vx[f1][2] - vx[f0][2]
		};

		const float side1[3]={
			vx[f2][0] - vx[f0][0],
			vx[f2][1] - vx[f0][1],
			vx[f2][2] - vx[f0][2]
		};

		const float normal[3]={
			side0[1]*side1[2] - side0[2]*side1[1],
			side0[2]*side1[0] - side0[0]*side1[2],
			side0[0]*side1[1] - side0[1]*side1[0]
		};

		float shade = normal[0]*lightdir[0] + normal[1]*lightdir[1] + normal[2]*lightdir[2];

		uint32_t alpha = VDRoundToInt((1.5f + 0.5f*shade) * 127.5f);

//		RenderTriangle((Pixel16 *)g_vbAboutDst.Address32(0,0), -g_vbAboutDst.pitch/2, tex, v+0, v+1, v+2, alpha);
//		RenderTriangle((Pixel16 *)g_vbAboutDst.Address32(0,0), -g_vbAboutDst.pitch/2, tex, v+2, v+1, v+3, alpha);
		RenderTriangle((Pixel16 *)pixels, info->stride / 2, tex, v+0, v+1, v+2, alpha);
		RenderTriangle((Pixel16 *)pixels, info->stride / 2, tex, v+2, v+1, v+3, alpha);
	}
//	InvalidateRect(hDlg, &rDirty2, FALSE);
}

extern "C" void Java_com_brobwind_cubic_CubicActivity_setTex(JNIEnv *env, jobject, jshortArray texture)
{
	int len = env->GetArrayLength(texture);
	short *ptr = env->GetShortArrayElements(texture, NULL);

	tex = new unsigned short[len];
	memcpy(tex, ptr, len * sizeof(unsigned short));

	env->ReleaseShortArrayElements(texture, ptr, 0);

	init();

#if 0
	unsigned short r = 0x00;
	unsigned short g = 0x00;
	unsigned short b = 0x00;

	for (int i = 0; i < 32; i++) {
		for (int j = 0; j < 32; j++) {
			if (i == 0 || j == 0 || i == 31 || j == 31) {
				*(tex + i * 32 + j) = r << 11 | g << 5 | b;
			}
		}
	}
#endif
}

extern "C" void Java_com_brobwind_cubic_CubicActivity_setBounce(JNIEnv * /*env*/, jobject,
		jint left, jint top, jint right, jint bottom)
{
	// Initialize cube vertices.
	int i;
	float rs;
	RECT r = RECT(left, top, right, bottom);

	if (r.right > r.bottom)
		rs = r.bottom / (3.6f*1.0f);
	else
		rs = r.right / (3.6f*1.0f);

	for(i=0; i<8; i++) {
		vx[i][0] = i&1 ? -rs : +rs;
		vx[i][1] = i&2 ? -rs : +rs;
		vx[i][2] = i&4 ? -rs : +rs;
	}

	cubeside = rs + rs;

	rBounce.left = rBounce.top = (int)ceil(rs*1.8);
	rBounce.right = r.right - rBounce.left;
	rBounce.bottom = r.bottom - rBounce.top;

	xpos = rand()%(rBounce.right-rBounce.left) + rBounce.left;
	ypos = rand()%(rBounce.bottom-rBounce.top) + rBounce.top;
	xvel = (rand()&2)-1;
	yvel = (rand()&2)-1;

	rDirtyLast.top = rDirtyLast.left = 0;
	rDirtyLast.right = rDirtyLast.bottom = 0;
}

extern "C" void Java_com_brobwind_cubic_CubicActivity_update(JNIEnv *env, jobject, jobject jbitmap)
{
	// Get the info.
	AndroidBitmapInfo info;
	int result = AndroidBitmap_getInfo(env, jbitmap, &info);
	if (result < 0) {
		throwIllegalStateException(env, (char*) "Cannot get bitmap info");
		return;
	}

	// Lock the pixels.
	void* pixels;
	result = AndroidBitmap_lockPixels(env, jbitmap, &pixels);
	if (result < 0) {
		throwIllegalStateException(env, (char*) "Cannot lock bitmap pixels");
		return;
	}

	// Update the pixels.
	updateBitmap(&info, pixels);

	// Unlock the pixels.
	result = AndroidBitmap_unlockPixels(env, jbitmap);
	if (result < 0) {
		throwIllegalStateException(env, (char*)"Cannot unlock bitmap pixels");
	}
}
